def add(M, a, b):
    if a not in M:
        M[a] = set()
    M[a].add(b)

    return M

#
# Converts set representation to multimap
#
def multimap(R):
    M = dict()
    for (x,y) in R:
        add(M, x, y)
    return M


def remove(M, a, b):
    if a in M and b in M[a]:
        M[a].remove(b)
        if M[a] == set():
            del M[a]
    return M

def map_element(M, a):
    if a not in M:
        return set()
    return M[a]


def inverse(M):
    N = dict()
    for x in M:
        for y in M[x]:
            add(N, y, x)
    return N

def compose(M, N):
    O = dict()
    for x in M:
        for y in M[x]:
            if y in N:
                for z in N[y]:
                    add(O, x, z)

    return O

def compose(M, N):
    return dict([(x, set([z for y in M[x] if y in N for z in N[y]])) for x in M if M[x] & set(N.keys()) != set()])

def merge(M, N):
    O = dict()
    for x in M:
        for y in M[x]:
            add(O, x, y)
    for x in N:
        for y in N[x]:
            add(O, x, y)
    return O

def merge(M, N):
    return dict([(x, M.get(x, set()) | N.get(x, set())) for x in set(M.keys()) | set(N.keys())])


if __name__ == '__main__':
    R = multimap(set([(1,'a'), (2,'b'), (1,'c'), (3,'a')]))
    assert add(R, 1, 'c') == R
    assert add(R, 2, 'c') == multimap(set([(1,'a'), (2,'b'), (1,'c'), (3,'a'), (2,'c')]))

    assert remove(R, 2, 'c') == multimap(set([(1,'a'), (2,'b'), (1,'c'), (3,'a')]))
    assert remove(R, 2, 'c') == R

    R = multimap(set([(1,'a'), (2,'b'), (1,'c'), (3,'a')]))
    assert map_element(R, 1) == set(['a', 'c'])
    assert map_element(R, 2) == set(['b'])
    assert map_element(R, 4) == set([])

    assert inverse(R) == multimap(set([('a',1), ('b',2), ('c',1), ('a',3)]))

    P = multimap(set([('a',2), ('c',3)]))
    
    assert compose(R, P) == multimap(set([(1,2), (1,3), (3,2)]))

    assert merge(inverse(R), P) == multimap(set([('a',1), ('b',2), ('c',1), ('a',3), ('a',2), ('c',3)]))
